/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#ifndef ATEMU_DECODE_TABLES_H
#define ATEMU_DECODE_TABLES_H

#include "opcodes.h"

static error_t ( *instruction_table[256] )( atmega_chip * ) = {
	atmega_op_nop,			/* 0x00 */
	atmega_op_movw,			/* 0x01 */
	atmega_op_muls,			/* 0x02 */
	atmega_op_mul_group,		/* 0x03 */
	atmega_op_cpc,			/* 0x04 */
	atmega_op_cpc,			/* 0x05 */
	atmega_op_cpc,			/* 0x06 */
	atmega_op_cpc,			/* 0x07 */
	atmega_op_sbc,			/* 0x08 */
	atmega_op_sbc,			/* 0x09 */
	atmega_op_sbc,			/* 0x0a */
	atmega_op_sbc,			/* 0x0b */
	atmega_op_add,			/* 0x0c */
	atmega_op_add,			/* 0x0d */
	atmega_op_add,			/* 0x0e */
	atmega_op_add,			/* 0x0f */

        atmega_op_cpse,			/* 0x10 */
        atmega_op_cpse,			/* 0x11 */
        atmega_op_cpse,			/* 0x12 */
        atmega_op_cpse,			/* 0x13 */
        atmega_op_cp,			/* 0x14 */
        atmega_op_cp,			/* 0x15 */
        atmega_op_cp,			/* 0x16 */
        atmega_op_cp,			/* 0x17 */
        atmega_op_sub,			/* 0x18 */
        atmega_op_sub,			/* 0x19 */
        atmega_op_sub,			/* 0x1a */
        atmega_op_sub,			/* 0x1b */
        atmega_op_adc,			/* 0x1c */
        atmega_op_adc,			/* 0x1d */
        atmega_op_adc,			/* 0x1e */
        atmega_op_adc,			/* 0x1f */

        atmega_op_and,			/* 0x20 */
        atmega_op_and,			/* 0x21 */
        atmega_op_and,			/* 0x22 */
        atmega_op_and,			/* 0x23 */
        atmega_op_eor,			/* 0x24 */
        atmega_op_eor,			/* 0x25 */
        atmega_op_eor,			/* 0x26 */
        atmega_op_eor,			/* 0x27 */
        atmega_op_or,			/* 0x28 */
        atmega_op_or,			/* 0x29 */
        atmega_op_or,			/* 0x2a */
        atmega_op_or,			/* 0x2b */
        atmega_op_mov,			/* 0x2c */
        atmega_op_mov,			/* 0x2d */
        atmega_op_mov,			/* 0x2e */
        atmega_op_mov,			/* 0x2f */

        atmega_op_cpi,			/* 0x30 */
        atmega_op_cpi,			/* 0x31 */
        atmega_op_cpi,			/* 0x32 */
        atmega_op_cpi,			/* 0x33 */
        atmega_op_cpi,			/* 0x34 */
        atmega_op_cpi,			/* 0x35 */
        atmega_op_cpi,			/* 0x36 */
        atmega_op_cpi,			/* 0x37 */
        atmega_op_cpi,			/* 0x38 */
        atmega_op_cpi,			/* 0x39 */
        atmega_op_cpi,			/* 0x3a */
        atmega_op_cpi,			/* 0x3b */
        atmega_op_cpi,			/* 0x3c */
        atmega_op_cpi,			/* 0x3d */
        atmega_op_cpi,			/* 0x3e */
        atmega_op_cpi,			/* 0x3f */

        atmega_op_sbci,			/* 0x40 */
        atmega_op_sbci,			/* 0x41 */
        atmega_op_sbci,			/* 0x42 */
        atmega_op_sbci,			/* 0x43 */
        atmega_op_sbci,			/* 0x44 */
        atmega_op_sbci,			/* 0x45 */
        atmega_op_sbci,			/* 0x46 */
        atmega_op_sbci,			/* 0x47 */
        atmega_op_sbci,			/* 0x48 */
        atmega_op_sbci,			/* 0x49 */
        atmega_op_sbci,			/* 0x4a */
        atmega_op_sbci,			/* 0x4b */
        atmega_op_sbci,			/* 0x4c */
        atmega_op_sbci,			/* 0x4d */
        atmega_op_sbci,			/* 0x4e */
        atmega_op_sbci,			/* 0x4f */

        atmega_op_subi,			/* 0x50 */
        atmega_op_subi,			/* 0x51 */
        atmega_op_subi,			/* 0x52 */
        atmega_op_subi,			/* 0x53 */
        atmega_op_subi,			/* 0x54 */
        atmega_op_subi,			/* 0x55 */
        atmega_op_subi,			/* 0x56 */
        atmega_op_subi,			/* 0x57 */
        atmega_op_subi,			/* 0x58 */
        atmega_op_subi,			/* 0x59 */
        atmega_op_subi,			/* 0x5a */
        atmega_op_subi,			/* 0x5b */
        atmega_op_subi,			/* 0x5c */
        atmega_op_subi,			/* 0x5d */
        atmega_op_subi,			/* 0x5e */
        atmega_op_subi,			/* 0x5f */

        atmega_op_ori,			/* 0x60 */
        atmega_op_ori,			/* 0x61 */
        atmega_op_ori,			/* 0x62 */
        atmega_op_ori,			/* 0x63 */
        atmega_op_ori,			/* 0x64 */
        atmega_op_ori,			/* 0x65 */
        atmega_op_ori,			/* 0x66 */
        atmega_op_ori,			/* 0x67 */
        atmega_op_ori,			/* 0x68 */
        atmega_op_ori,			/* 0x69 */
        atmega_op_ori,			/* 0x6a */
        atmega_op_ori,			/* 0x6b */
        atmega_op_ori,			/* 0x6c */
        atmega_op_ori,			/* 0x6d */
        atmega_op_ori,			/* 0x6e */
        atmega_op_ori,			/* 0x6f */

        atmega_op_andi,			/* 0x70 */
        atmega_op_andi,			/* 0x71 */
        atmega_op_andi,			/* 0x72 */
        atmega_op_andi,			/* 0x73 */
        atmega_op_andi,			/* 0x74 */
        atmega_op_andi,			/* 0x75 */
        atmega_op_andi,			/* 0x76 */
        atmega_op_andi,			/* 0x77 */
        atmega_op_andi,			/* 0x78 */
        atmega_op_andi,			/* 0x79 */
        atmega_op_andi,			/* 0x7a */
        atmega_op_andi,			/* 0x7b */
        atmega_op_andi,			/* 0x7c */
        atmega_op_andi,			/* 0x7d */
        atmega_op_andi,			/* 0x7e */
        atmega_op_andi,			/* 0x7f */

        atmega_op_ld_ldd,		/* 0x80 */
        atmega_op_ld_ldd,		/* 0x81 */
        atmega_op_st_std,		/* 0x82 */
        atmega_op_st_std,		/* 0x83 */
        atmega_op_ld_ldd,		/* 0x84 */
        atmega_op_ld_ldd,		/* 0x85 */
        atmega_op_st_std,		/* 0x86 */
        atmega_op_st_std,		/* 0x87 */
        atmega_op_ld_ldd,		/* 0x88 */
        atmega_op_ld_ldd,		/* 0x89 */
        atmega_op_st_std,		/* 0x8a */
        atmega_op_st_std,		/* 0x8b */
        atmega_op_ld_ldd,		/* 0x8c */
        atmega_op_ld_ldd,		/* 0x8d */
        atmega_op_st_std,		/* 0x8e */
        atmega_op_st_std,		/* 0x8f */

	atmega_op_ld_elpm_pop,		/* 0x90 */
	atmega_op_ld_elpm_pop,		/* 0x91 */
	atmega_op_st_push,		/* 0x92 */
	atmega_op_st_push,		/* 0x93 */
	atmega_op_flags_jump_arith,	/* 0x94 */
	atmega_op_flags_jump_arith,	/* 0x95 */
	atmega_op_adiw,			/* 0x96 */
	atmega_op_sbiw,			/* 0x97 */
	atmega_op_cbi,			/* 0x98 */
	atmega_op_sbic,			/* 0x99 */
	atmega_op_sbi,			/* 0x9a */
	atmega_op_sbis,			/* 0x9b */
	atmega_op_mul,			/* 0x9c */
	atmega_op_mul,			/* 0x9d */
	atmega_op_mul,			/* 0x9e */
	atmega_op_mul,			/* 0x9f */

        atmega_op_ld_ldd,		/* 0xa0 */
        atmega_op_ld_ldd,		/* 0xa1 */
        atmega_op_st_std,		/* 0xa2 */
        atmega_op_st_std,		/* 0xa3 */
        atmega_op_ld_ldd,		/* 0xa4 */
        atmega_op_ld_ldd,		/* 0xa5 */
        atmega_op_st_std,		/* 0xa6 */
        atmega_op_st_std,		/* 0xa7 */
        atmega_op_ld_ldd,		/* 0xa8 */
        atmega_op_ld_ldd,		/* 0xa9 */
        atmega_op_st_std,		/* 0xaa */
        atmega_op_st_std,		/* 0xab */
        atmega_op_ld_ldd,		/* 0xac */
        atmega_op_ld_ldd,		/* 0xad */
        atmega_op_st_std,		/* 0xae */
        atmega_op_st_std,		/* 0xaf */

	atmega_op_in,			/* 0xb0 */
	atmega_op_in,			/* 0xb1 */
	atmega_op_in,			/* 0xb2 */
	atmega_op_in,			/* 0xb3 */
	atmega_op_in,			/* 0xb4 */
	atmega_op_in,			/* 0xb5 */
	atmega_op_in,			/* 0xb6 */
	atmega_op_in,			/* 0xb7 */
	atmega_op_out,			/* 0xb8 */
	atmega_op_out,			/* 0xb9 */
	atmega_op_out,			/* 0xba */
	atmega_op_out,			/* 0xbb */
	atmega_op_out,			/* 0xbc */
	atmega_op_out,			/* 0xbd */
	atmega_op_out,			/* 0xbe */
	atmega_op_out,			/* 0xbf */

	atmega_op_rjmp,			/* 0xc0 */
	atmega_op_rjmp,			/* 0xc1 */
	atmega_op_rjmp,			/* 0xc2 */
	atmega_op_rjmp,			/* 0xc3 */
	atmega_op_rjmp,			/* 0xc4 */
	atmega_op_rjmp,			/* 0xc5 */
	atmega_op_rjmp,			/* 0xc6 */
	atmega_op_rjmp,			/* 0xc7 */
	atmega_op_rjmp,			/* 0xc8 */
	atmega_op_rjmp,			/* 0xc9 */
	atmega_op_rjmp,			/* 0xca */
	atmega_op_rjmp,			/* 0xcb */
	atmega_op_rjmp,			/* 0xcc */
	atmega_op_rjmp,			/* 0xcd */
	atmega_op_rjmp,			/* 0xce */
	atmega_op_rjmp,			/* 0xcf */

	atmega_op_rcall,		/* 0xd0 */
	atmega_op_rcall,		/* 0xd1 */
	atmega_op_rcall,		/* 0xd2 */
	atmega_op_rcall,		/* 0xd3 */
	atmega_op_rcall,		/* 0xd4 */
	atmega_op_rcall,		/* 0xd5 */
	atmega_op_rcall,		/* 0xd6 */
	atmega_op_rcall,		/* 0xd7 */
	atmega_op_rcall,		/* 0xd8 */
	atmega_op_rcall,		/* 0xd9 */
	atmega_op_rcall,		/* 0xda */
	atmega_op_rcall,		/* 0xdb */
	atmega_op_rcall,		/* 0xdc */
	atmega_op_rcall,		/* 0xdd */
	atmega_op_rcall,		/* 0xde */
	atmega_op_rcall,		/* 0xdf */

	atmega_op_ldi,			/* 0xe0 */
	atmega_op_ldi,			/* 0xe1 */
	atmega_op_ldi,			/* 0xe2 */
	atmega_op_ldi,			/* 0xe3 */
	atmega_op_ldi,			/* 0xe4 */
	atmega_op_ldi,			/* 0xe5 */
	atmega_op_ldi,			/* 0xe6 */
	atmega_op_ldi,			/* 0xe7 */
	atmega_op_ldi,			/* 0xe8 */
	atmega_op_ldi,			/* 0xe9 */
	atmega_op_ldi,			/* 0xea */
	atmega_op_ldi,			/* 0xeb */
	atmega_op_ldi,			/* 0xec */
	atmega_op_ldi,			/* 0xed */
	atmega_op_ldi,			/* 0xee */
	atmega_op_ldi,			/* 0xef */

	atmega_op_brbs,			/* 0xf0 */
	atmega_op_brbs,			/* 0xf1 */
	atmega_op_brbs,			/* 0xf2 */
	atmega_op_brbs,			/* 0xf3 */
	atmega_op_brbc,			/* 0xf4 */
	atmega_op_brbc,			/* 0xf5 */
	atmega_op_brbc,			/* 0xf6 */
	atmega_op_brbc,			/* 0xf7 */
	atmega_op_bld,			/* 0xf8 */
	atmega_op_bld,			/* 0xf9 */
	atmega_op_bst,			/* 0xfa */
	atmega_op_bst,			/* 0xfb */
	atmega_op_sbrc,			/* 0xfc */
	atmega_op_sbrc,			/* 0xfd */
	atmega_op_sbrs,			/* 0xfe */
	atmega_op_sbrs,			/* 0xff */
};

/**
 * Cryptic? Yes! Fun? Yes!
 *
 * 0 : none
 * 1 : ---- ---- -aaa -bbb
 * 2 : ---- ---- aaaa bbbb
 * 3 : ---- --ba aaaa bbbb
 * 4 : ---- kkkk aaaa kkkk
 * 5 : ---- ---a aaaa ----
 * 6 : --q- qq-a aaaa -qqq
 * 7 : ---- ---a aaaa ---- kkkk kkkk kkkk kkkk
 * 8 : ---- ---- -bbb ---- or (5) and check for (7)
 * 9 : (2) but fill register index from most sig. bit down...
 * 10: ---- ---- kkaa kkkk
 * 11: ---- ---- aaaa abbb
 * 12: ---- -bba aaaa bbbb
 * 13: ---- kkkk kkkk kkkk
 * 14: ---- --kk kkkk ksss
 * 15: ---- ---a aaaa -sss
 *
 **/

static uint8_t encode_table[256] = {
	0, 9, 2, 1, 3, 3, 3, 3,   /* 0x0 */
	3, 3, 3, 3, 3, 3, 3, 3,

	3, 3, 3, 3, 3, 3, 3, 3,   /* 0x1 */
	3, 3, 3, 3, 3, 3, 3, 3,

	3, 3, 3, 3, 3, 3, 3, 3,   /* 0x2 */
	3, 3, 3, 3, 3, 3, 3, 3,

	4, 4, 4, 4, 4, 4, 4, 4,   /* 0x3 */
	4, 4, 4, 4, 4, 4, 4, 4,

	4, 4, 4, 4, 4, 4, 4, 4,   /* 0x4 */
	4, 4, 4, 4, 4, 4, 4, 4,

	4, 4, 4, 4, 4, 4, 4, 4,   /* 0x5 */
	4, 4, 4, 4, 4, 4, 4, 4,

	4, 4, 4, 4, 4, 4, 4, 4,   /* 0x6 */
	4, 4, 4, 4, 4, 4, 4, 4,

	4, 4, 4, 4, 4, 4, 4, 4,   /* 0x7 */
	4, 4, 4, 4, 4, 4, 4, 4,

	6, 6, 6, 6, 6, 6, 6, 6,   /* 0x8 */
	6, 6, 6, 6, 6, 6, 6, 6,

	7, 7, 7, 7, 8, 8, 10, 10,   /* 0x9 */
	11, 11, 11, 11, 3, 3, 3, 3,

	6, 6, 6, 6, 6, 6, 6, 6,   /* 0xa */
	6, 6, 6, 6, 6, 6, 6, 6,

	12, 12, 12, 12, 12, 12, 12, 12,   /* 0xb */
	12, 12, 12, 12, 12, 12, 12, 12,

	13, 13, 13, 13, 13, 13, 13, 13,   /* 0xc */
	13, 13, 13, 13, 13, 13, 13, 13,

	13, 13, 13, 13, 13, 13, 13, 13,   /* 0xd */
	13, 13, 13, 13, 13, 13, 13, 13,

	4, 4, 4, 4, 4, 4, 4, 4,   /* 0xe */
	4, 4, 4, 4, 4, 4, 4, 4,

	14, 14, 14, 14, 14, 14, 14, 14,   /* 0xf */
	15, 15, 15, 15, 15, 15, 15, 15
};

#endif
