/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>
#include <libgen.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include "atemu.h"
#include "chip_data.h"
#include "chip_data_struct.h"
#include "breakpoints.h"
#include "watchpoints.h"

#include "license.h"
#include "version.h"

int watchpoint_chip_sel;
int breakpoint_chip_sel;
static GtkWidget *watch_popup = NULL;
static GtkWidget *popup = NULL;
static GtkWidget *about_popup = NULL;
static char *openpath = NULL;

void
on_new1_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}

static void file_ok_sel( GtkWidget *w, GtkFileSelection *fs )
{
    const char *filename;
    char *filecopy;
    char *filepath;

    filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (fs));

    /* maintain file->open path accross mutiple opens */
    filecopy = strdup(filename);
    filepath = dirname(filecopy);
    if(openpath != NULL){
        free(openpath);
    }
    /* length of other path, +1 for '/' +1 for '\0' */
    openpath = malloc(strlen(filepath) + 2);
    sprintf(openpath, "%s/", filepath);
        
    atemu_quit();

    atemu_init( filename );
    
    gtk_widget_destroy(GTK_WIDGET(fs));
}

void
on_open1_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    GtkWidget *filew;

    if(openpath == NULL){
        openpath = strdup(DATADIR"/");
    }

    filew = gtk_file_selection_new ("Please select a file.");
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(filew), openpath);
    gtk_file_selection_set_select_multiple(GTK_FILE_SELECTION(filew), FALSE);

    gtk_widget_show(filew);
    /* Connect the ok_button to file_ok_sel function */
    g_signal_connect (G_OBJECT (GTK_FILE_SELECTION (filew)->ok_button),
            "clicked", G_CALLBACK (file_ok_sel), (gpointer) filew);

    /* Connect the cancel_button to destroy the widget */
    g_signal_connect_swapped (G_OBJECT (GTK_FILE_SELECTION (filew)->cancel_button),
            "clicked", G_CALLBACK (gtk_widget_destroy),
            G_OBJECT (filew));
}


void
on_save1_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


void
on_save_as1_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


void
on_quit1_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    atemu_quit();
    gtk_main_quit( );
}


void
on_cut1_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


void
on_copy1_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


void
on_paste1_activate                     (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


void
on_delete1_activate                    (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}

void
close_about_popup( )
{
    if ( about_popup != NULL )
    {
        gtk_widget_destroy( about_popup );
        about_popup = NULL;
    }
}

void
on_about1_activate                     (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
    GtkWidget *vbox;
    GtkWidget *label;
    GtkWidget *license;
    GtkWidget *scroll;
    GtkWidget *box1, *button_close;

    GtkTextBuffer *lbuff;
    GtkWidget *license_frame;
    GtkTextIter iter;
    
    /*stop more than one window from opening*/
    if( about_popup != NULL )
    {
        return;
    }
    
    about_popup = gtk_window_new( GTK_WINDOW_TOPLEVEL );
    gtk_window_set_default_size( GTK_WINDOW( about_popup ), -1, 360 );
    gtk_window_set_title( GTK_WINDOW( about_popup ), "About xatdb" );
    gtk_container_set_border_width (GTK_CONTAINER( about_popup ), 10);
    
    vbox = gtk_vbox_new( FALSE, 5 );
    
    gtk_container_add( GTK_CONTAINER( about_popup ), vbox );
    
    label = gtk_label_new( XATDB_VERSION );
    gtk_label_set_selectable( GTK_LABEL( label ), TRUE );
    gtk_box_pack_start( GTK_BOX( vbox ), label, FALSE, FALSE, 2 );
    label = gtk_label_new( "By:\tJonathan Polley \n\tDion Blazakis\n\tDaniel Rusk\n\tJonathan McGee\n\tManish Karir\n" );
    gtk_label_set_selectable( GTK_LABEL( label ), TRUE );
    gtk_box_pack_start( GTK_BOX( vbox ), label, FALSE, FALSE, 2 );
    label = gtk_label_new( "http://www.isr.umd.edu/CSHCN/research/atemu/" );
    gtk_label_set_selectable( GTK_LABEL( label ), TRUE );
    gtk_box_pack_start( GTK_BOX( vbox ), label, FALSE, FALSE, 2 );

    license_frame = gtk_frame_new( NULL );
    gtk_frame_set_shadow_type( GTK_FRAME( license_frame ), GTK_SHADOW_IN );

    scroll = gtk_scrolled_window_new( NULL, NULL );
    gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW( scroll ),
        GTK_POLICY_NEVER, GTK_POLICY_ALWAYS );

    license = gtk_text_view_new();
    lbuff = gtk_text_view_get_buffer( GTK_TEXT_VIEW( license ) );
    gtk_text_buffer_create_tag(lbuff, "monospace", 
            "family", "courier", 
            NULL);
    gtk_text_buffer_get_start_iter(lbuff, &iter);
    gtk_text_buffer_insert_with_tags_by_name( lbuff, &iter, license_text,
        strlen( license_text ), "monospace", NULL );
    gtk_text_view_set_editable( GTK_TEXT_VIEW( license ), FALSE );

    gtk_container_add( GTK_CONTAINER( scroll ), license );
    gtk_container_add( GTK_CONTAINER( license_frame ), scroll );
    gtk_box_pack_start( GTK_BOX( vbox ), license_frame, TRUE, TRUE, 2 );
    
    box1 = gtk_hbutton_box_new( );
    
    button_close = gtk_button_new_from_stock( GTK_STOCK_OK );
    gtk_button_box_set_layout( GTK_BUTTON_BOX( box1 ), 
                                GTK_BUTTONBOX_END );
     
    gtk_box_pack_end( GTK_BOX( box1 ), button_close, FALSE, FALSE, 0 );
    /* Connect the close button to destroy the widget */
    g_signal_connect_swapped (G_OBJECT( button_close ),
            "clicked", G_CALLBACK( close_about_popup ),
            NULL );
    gtk_box_pack_end( GTK_BOX( vbox ), box1, FALSE, FALSE, 3 );

    gtk_widget_show_all( about_popup );
}

void
button_step_clicked                    (GtkWidget *widget,
                                        gpointer   data)
{
    atemu_step();
}

void
button_cstep_clicked                      (GtkWidget *widget,
                                        gpointer   data)
{
    activate_auto_breakpoints(0);
    toolbar_sense_go();
    atemu_go();
}

void
button_go_clicked                      (GtkWidget *widget,
                                        gpointer   data)
{
    deactivate_auto_breakpoints(0);
    toolbar_sense_go();
    atemu_go();
}

void
button_stop_clicked                    (GtkWidget *widget,
                                        gpointer   data)
{
    toolbar_sense_stop();
    atemu_stop();
}

void
button_restart_clicked                 (GtkWidget *widget,
                                        gpointer   data)
{
    atemu_quit();

    atemu_init( NULL );
}

void close_watch_popup( )
{
    if ( watch_popup != NULL )
    {
        gtk_widget_destroy( watch_popup );
        watch_popup = NULL;
    }
}

void close_break_popup( )
{
    if ( popup != NULL )
    {
        gtk_widget_destroy( popup );
        popup = NULL;
    }
}

void
close_popups( )
{
    close_about_popup();
    close_watch_popup();
    close_break_popup();
}

void
button_set_break_clicked               (GtkWidget *widget,
                                        gpointer   data)
{
    GtkWidget *chips_menu_item, *entry, *button_apply, 
        *button_close, *label1,
        *label2, *table, *hbox1, 
        *image, *label, *chips_menu, *chips_option_menu;
    char buf[128];
    int i;
    
    /*stop more than one window from opening*/
    if( popup != NULL )
    {
        return;
    }
    
    if( chip_array_cnt == 0 )
    {
        printf("no chips\n");
        return;
    }
   
    breakpoint_chip_sel = curr_chip + 1;
    
    popup = gtk_window_new( GTK_WINDOW_TOPLEVEL );
    gtk_window_set_title( GTK_WINDOW( popup ), "Set Breakpoint" );
    gtk_container_set_border_width (GTK_CONTAINER( popup ), 10);
    
    table = gtk_table_new( 3, 4, TRUE );
    gtk_container_add( GTK_CONTAINER( popup ), table );
    
    label1 = gtk_label_new( "Chip ID" );
    label2 = gtk_label_new( "Breakpoint Address" );
    gtk_table_attach_defaults( GTK_TABLE( table ), label1, 0, 2, 0, 1 );
    gtk_table_attach_defaults( GTK_TABLE( table ), label2, 2, 4, 0, 1 );

    chips_menu = gtk_menu_new( );
    chips_menu_item = gtk_menu_item_new_with_mnemonic( "All Chips" );
    gtk_menu_shell_append( GTK_MENU_SHELL( chips_menu ), chips_menu_item );
    
    for( i = 0; i < chip_array_cnt; i++ )
    {
        snprintf( buf, sizeof(buf), "Chip %d", i );
        buf[sizeof(buf)-1] = '\n';
        chips_menu_item = gtk_menu_item_new_with_mnemonic( buf );
        gtk_menu_shell_append( GTK_MENU_SHELL( chips_menu ), chips_menu_item );
    }

    chips_option_menu = gtk_option_menu_new( );
    gtk_option_menu_set_menu( GTK_OPTION_MENU( chips_option_menu ), chips_menu );
    
    gtk_option_menu_set_history( GTK_OPTION_MENU( chips_option_menu ), curr_chip + 1 );
    
    /*connect the option menu to the change_breakpoint_chip function*/
    g_signal_connect( G_OBJECT( chips_option_menu ), "changed", 
            G_CALLBACK( change_breakpoint_chip ), (gpointer) chips_option_menu );
    
    entry = gtk_entry_new( );
    /*connect the text entry to the apply_breakpoint function*/
    g_signal_connect( G_OBJECT( entry ), "activate", 
            G_CALLBACK( apply_breakpoint ), (gpointer) entry );

    gtk_table_attach_defaults( GTK_TABLE( table ), chips_option_menu, 0, 2, 1, 2 );
    gtk_table_attach_defaults( GTK_TABLE( table ), entry, 2, 4, 1, 2 );
    
    hbox1 = gtk_hbox_new( FALSE, 0 );
    image = gtk_image_new_from_stock( GTK_STOCK_ADD, GTK_ICON_SIZE_BUTTON );
    label = gtk_label_new( "Add" );
    gtk_box_pack_start( GTK_BOX( hbox1 ), image, FALSE, FALSE, 3 );
    gtk_box_pack_start( GTK_BOX( hbox1 ), label, FALSE, FALSE, 3 );
    button_apply = gtk_button_new( );
    gtk_container_add( GTK_CONTAINER( button_apply ), hbox1 );
    
    /*connect the apply button to the apply_breakpoint function*/ 
    g_signal_connect( G_OBJECT( button_apply ), "clicked",
            G_CALLBACK( apply_breakpoint ), (gpointer) entry );
    gtk_table_attach_defaults( GTK_TABLE( table ), button_apply, 2, 3, 2, 3 );

    hbox1 = gtk_hbox_new( FALSE, 0 );
    button_close = gtk_button_new( );
    image = gtk_image_new_from_stock( GTK_STOCK_CANCEL, GTK_ICON_SIZE_BUTTON );
    label = gtk_label_new( "Cancel" );
    gtk_box_pack_start( GTK_BOX( hbox1 ), image, FALSE, FALSE, 3 );
    gtk_box_pack_start( GTK_BOX( hbox1 ), label, FALSE, FALSE, 3 );
    gtk_container_add( GTK_CONTAINER( button_close ), hbox1 );

    /* Connect the close button to destroy the widget */
    g_signal_connect_swapped (G_OBJECT( button_close ),
            "clicked", G_CALLBACK( close_break_popup ),
            NULL );
    gtk_table_attach_defaults( GTK_TABLE( table ), button_close, 3, 4, 2, 3 );
    
    gtk_widget_show_all( popup );
    
    /*gtk_widget_grab_default( button_apply );*/
    gtk_widget_grab_focus( entry );
}
  
void 
apply_breakpoint                       (GtkWidget *widget,
                                        gpointer   data)
{
    GtkWidget *dialog;
    int line;
    int i;
    long addr;

    if( sscanf( (char *) gtk_entry_get_text( GTK_ENTRY( data ) ), 
            "%lx", &addr ) <= 0 )
    {
        dialog = gtk_message_dialog_new( GTK_WINDOW( popup ),
                GTK_DIALOG_DESTROY_WITH_PARENT,
                GTK_MESSAGE_ERROR,
                GTK_BUTTONS_CLOSE,
                "Invalid Address '%s'",
                (char *) gtk_entry_get_text( GTK_ENTRY( data ) ) );
        /* Destroy the dialog when the user responds to it (e.g. clicks a button) */
        g_signal_connect_swapped (GTK_OBJECT (dialog), "response",
                G_CALLBACK (gtk_widget_destroy),
                GTK_OBJECT (dialog));

        gtk_widget_show( dialog );
        return;
    }

    /*apply to all chips*/
    if( breakpoint_chip_sel == 0 )
    {
        for( i = 0; i < chip_array_cnt; i++ )
        {
            line = address_to_text_line( i, addr );
    
            if( line == -1 )            
            {
                dialog = gtk_message_dialog_new( GTK_WINDOW( watch_popup ),
                        GTK_DIALOG_DESTROY_WITH_PARENT,
                        GTK_MESSAGE_ERROR,
                        GTK_BUTTONS_CLOSE,
                        "Invalid Address '%s'",
                        (char *) gtk_entry_get_text( GTK_ENTRY( data ) ) );
                /* Destroy the dialog when the user responds to it (e.g. clicks a button) */
                g_signal_connect_swapped (GTK_OBJECT (dialog), "response",
                        G_CALLBACK (gtk_widget_destroy),
                        GTK_OBJECT (dialog));
                
                gtk_widget_show( dialog );
                
                return;
            }
            
            add_asm_breakpoint( i, addr );
        }
    }
    else /*specific chip selected*/
    {
        i = breakpoint_chip_sel - 1;
        line = address_to_text_line( i, addr );
        
        if( line == -1 )
        {
            dialog = gtk_message_dialog_new( GTK_WINDOW( watch_popup ),
                    GTK_DIALOG_DESTROY_WITH_PARENT,
                    GTK_MESSAGE_ERROR,
                    GTK_BUTTONS_CLOSE,
                    "Invalid Address '%s'",
                    (char *) gtk_entry_get_text( GTK_ENTRY( data ) ) );
            /* Destroy the dialog when the user responds to it (e.g. clicks a button) */
            g_signal_connect_swapped (GTK_OBJECT (dialog), "response",
                    G_CALLBACK (gtk_widget_destroy),
                    GTK_OBJECT (dialog));
            gtk_widget_show( dialog );
            return;
        }
    
        add_asm_breakpoint( i, addr );
    }
    close_break_popup();
}

void
change_breakpoint_chip                 (GtkWidget *widget,
                                        gpointer   data)
{
    breakpoint_chip_sel = gtk_option_menu_get_history ( GTK_OPTION_MENU( data ) );
}

void
change_watchpoint_chip                 (GtkWidget *widget,
                                        gpointer   data)
{
    watchpoint_chip_sel = gtk_option_menu_get_history ( GTK_OPTION_MENU( data ) );
}

void 
apply_watchpoint                       (GtkWidget *widget,
                                        gpointer   data)
{
    GtkWidget *dialog;
    int i;
    long addr;
    
    
    if( sscanf( (char *) gtk_entry_get_text( GTK_ENTRY( data ) ), 
            "%lx", &addr ) <= 0 )
    {
        dialog = gtk_message_dialog_new( GTK_WINDOW( watch_popup ),
                GTK_DIALOG_DESTROY_WITH_PARENT,
                GTK_MESSAGE_ERROR,
                GTK_BUTTONS_CLOSE,
                "Invalid Address '%s'",
                (char *) gtk_entry_get_text( GTK_ENTRY( data ) ) );
        /* Destroy the dialog when the user responds to it (e.g. clicks a button) */
        g_signal_connect_swapped (GTK_OBJECT (dialog), "response",
                G_CALLBACK (gtk_widget_destroy),
                GTK_OBJECT (dialog));

        gtk_widget_show( dialog );
        return;
    }

    /*apply to all chips*/
    if( watchpoint_chip_sel == 0 )
    {
        for( i = 0; i < chip_array_cnt; i++ )
        {
            add_watchpoint( i, addr );
        }
    }
    else /*specific chip selected*/
    {
        add_watchpoint( watchpoint_chip_sel - 1, addr );
    }
    
    close_watch_popup();
}

void
button_set_watch_clicked               (GtkWidget *widget,
                                        gpointer   data)
{

    GtkWidget *chips_menu_item, *entry, *button_apply, 
        *button_close, *label1,
        *label2, *table, *hbox1, 
        *image, *label, *chips_menu, *chips_option_menu;
    char buf[128];
    int i;
    
    /*avoid multiple popups*/
    if( watch_popup != NULL )
    {
        return;
    }
    
    if( chip_array_cnt == 0 )
    {
        printf("no chips\n");
        return;
    }
   
    watchpoint_chip_sel = curr_chip + 1;
    
    watch_popup = gtk_window_new( GTK_WINDOW_TOPLEVEL );
    gtk_window_set_title( GTK_WINDOW( watch_popup ), "Set Watchpoint" );
    gtk_container_set_border_width (GTK_CONTAINER( watch_popup ), 10);
    
    table = gtk_table_new( 3, 4, TRUE );
    gtk_container_add( GTK_CONTAINER( watch_popup ), table );
    
    label1 = gtk_label_new( "Chip ID" );
    label2 = gtk_label_new( "Watchpoint Address" );
    gtk_table_attach_defaults( GTK_TABLE( table ), label1, 0, 2, 0, 1 );
    gtk_table_attach_defaults( GTK_TABLE( table ), label2, 2, 4, 0, 1 );

    chips_menu = gtk_menu_new( );
    chips_menu_item = gtk_menu_item_new_with_mnemonic( "All Chips" );
    gtk_menu_shell_append( GTK_MENU_SHELL( chips_menu ), chips_menu_item );
    
    for( i = 0; i < chip_array_cnt; i++ )
    {
        snprintf( buf, sizeof(buf), "Chip %d", i );
        buf[sizeof(buf)-1] = '\n';
        chips_menu_item = gtk_menu_item_new_with_mnemonic( buf );
        gtk_menu_shell_append( GTK_MENU_SHELL( chips_menu ), chips_menu_item );
    }

    chips_option_menu = gtk_option_menu_new( );
    gtk_option_menu_set_menu( GTK_OPTION_MENU( chips_option_menu ), chips_menu );
    
    gtk_option_menu_set_history( GTK_OPTION_MENU( chips_option_menu ), curr_chip + 1 );
    
    /*connect the option menu to the change_watchpoint_chip function*/
    g_signal_connect( G_OBJECT( chips_option_menu ), "changed", 
            G_CALLBACK( change_watchpoint_chip ), (gpointer) chips_option_menu );
    
    entry = gtk_entry_new( );
    /*connect the text entry to the apply_watchpoint function*/
    g_signal_connect( G_OBJECT( entry ), "activate", 
            G_CALLBACK( apply_watchpoint ), (gpointer) entry );

    gtk_table_attach_defaults( GTK_TABLE( table ), chips_option_menu, 0, 2, 1, 2 );
    gtk_table_attach_defaults( GTK_TABLE( table ), entry, 2, 4, 1, 2 );
    
    hbox1 = gtk_hbox_new( FALSE, 0 );
    image = gtk_image_new_from_stock( GTK_STOCK_ADD, GTK_ICON_SIZE_BUTTON );
    label = gtk_label_new( "Add" );
    gtk_box_pack_start( GTK_BOX( hbox1 ), image, FALSE, FALSE, 3 );
    gtk_box_pack_start( GTK_BOX( hbox1 ), label, FALSE, FALSE, 3 );
    button_apply = gtk_button_new( );
    gtk_container_add( GTK_CONTAINER( button_apply ), hbox1 );
    
    /*connect the apply button to the apply_watchpoint function*/ 
    g_signal_connect( G_OBJECT( button_apply ), "clicked",
            G_CALLBACK( apply_watchpoint ), (gpointer) entry );
    gtk_table_attach_defaults( GTK_TABLE( table ), button_apply, 2, 3, 2, 3 );

    hbox1 = gtk_hbox_new( FALSE, 0 );
    button_close = gtk_button_new( );
    image = gtk_image_new_from_stock( GTK_STOCK_CANCEL, GTK_ICON_SIZE_BUTTON );
    label = gtk_label_new( "Cancel" );
    gtk_box_pack_start( GTK_BOX( hbox1 ), image, FALSE, FALSE, 3 );
    gtk_box_pack_start( GTK_BOX( hbox1 ), label, FALSE, FALSE, 3 );
    gtk_container_add( GTK_CONTAINER( button_close ), hbox1 );

    /* Connect the close button to destroy the widget */
    g_signal_connect_swapped (G_OBJECT( button_close ),
            "clicked", G_CALLBACK( close_watch_popup ),
            NULL);
    gtk_table_attach_defaults( GTK_TABLE( table ), button_close, 3, 4, 2, 3 );
    
    gtk_widget_show_all( watch_popup );
    
    /*gtk_widget_grab_default( button_apply );*/
    gtk_widget_grab_focus( entry );
}

gboolean on_window1_configure(GtkWidget *widget, GdkEventConfigure *event){
    int vpos, hpos;
    int i;

    if(vpane_ratio < 0){
        vpane_ratio = -vpane_ratio;
    }

    if(hpane_ratio < 0){
        hpane_ratio = -hpane_ratio;
    }

    hpos = (int)(hpane_ratio * ((double)event->width));
    vpos = (int)(vpane_ratio * ((double)event->height));

    for(i=0; i<chip_array_cnt; i++){
        gtk_paned_set_position(GTK_PANED (chip_array[i].hpaned), hpos);
        gtk_paned_set_position(GTK_PANED (chip_array[i].vpaned), vpos);
    }

    return FALSE;
}

void vpane_change_position(GObject *object, GParamSpec *pspec, gpointer data){
    if(vpane_ratio >= 0){
        int x, y;
        int pos;

        gtk_window_get_size(GTK_WINDOW(window1), &x, &y);
        g_object_get(object, pspec->name, &pos, NULL);

        vpane_ratio = ((double)pos) / ((double)y);
    }
}

void hpane_change_position(GObject *object, GParamSpec *pspec, gpointer data){
    if(hpane_ratio >= 0){
        int x, y;
        int pos;
//        int i;

        gtk_window_get_size(GTK_WINDOW(window1), &x, &y);
        g_object_get(object, pspec->name, &pos, NULL);

        hpane_ratio = ((double)pos) / ((double)x);

//        for(i=0; i<chip_array_cnt; i++){
//            printf("%d handlers removed.\n", g_signal_handlers_disconnect_by_func(chip_array[i].hpaned, hpane_change_position, NULL));
//            g_signal_stop_emission_by_name(chip_array[i].hpaned, "notify::position");
//            gtk_paned_set_position(GTK_PANED (chip_array[i].hpaned), pos);
//            g_signal_connect(G_OBJECT(chip_array[i].hpaned), "notify::position", G_CALLBACK(hpane_change_position), NULL);
//        }
    }
}

void notebook_chips_select(GtkNotebook *book, GtkNotebookPage *page, guint page_num, gpointer data){
    int x, y;
    int hpos, vpos;
    
    curr_chip = page_num - 1;

    if(curr_chip >= 0){
        gtk_window_get_size(GTK_WINDOW(window1), &x, &y);
        hpos = (int)(hpane_ratio * ((double)x));
        vpos = (int)(vpane_ratio * ((double)y));

        gtk_paned_set_position(GTK_PANED (chip_array[curr_chip].hpaned), hpos);
        gtk_paned_set_position(GTK_PANED (chip_array[curr_chip].vpaned), vpos);
    }
}

void source_popup_add_break(GtkMenuItem *menuitem, gpointer user_data){
    GtkTextMark *mark;
    GtkTextIter iter;
    char *cfile;
    GtkTextBuffer *buff;
    int line;
    long addr;

    buff = gtk_text_view_get_buffer(GTK_TEXT_VIEW(chip_array[curr_chip].textview_source));
    mark = gtk_text_buffer_get_insert(buff);
    gtk_text_buffer_get_iter_at_mark(buff, &iter, mark);
    line = gtk_text_iter_get_line(&iter) + 1;
    cfile = cfile_index_to_cfile(curr_chip, gtk_option_menu_get_history(GTK_OPTION_MENU(chip_array[curr_chip].cfile_select)));
    addr = c_line_to_address(curr_chip, cfile, line);

    add_source_breakpoint(curr_chip, addr, cfile);
}

void code_popup_add_break(GtkMenuItem *menuitem, gpointer user_data){
    GtkTextMark *mark;
    GtkTextIter iter;
    int line;
    long addr;

    mark = gtk_text_buffer_get_insert(chip_array[curr_chip].code);
    gtk_text_buffer_get_iter_at_mark(chip_array[curr_chip].code, &iter, mark);
    line = gtk_text_iter_get_line(&iter);
    addr = text_line_to_address(curr_chip, line);

    add_asm_breakpoint(curr_chip, addr);
}

void watch_popup_clear_watch(GtkWidget *widget, gpointer data){
    GtkTreeIter iter;
    GtkTreePath *path;
    GtkTreeViewColumn *column;
    long addr;

    /*find out where the cursor is and highlight the row*/
    gtk_tree_view_get_cursor( (GTK_TREE_VIEW(chip_array[curr_chip].treeview_watch)), 
        &path, &column );
    gtk_tree_view_row_activated ( GTK_TREE_VIEW( chip_array[curr_chip].treeview_watch ),
         path, column );

    /*if the iter does not exist, then return*/
    if( !gtk_tree_model_get_iter( 
                GTK_TREE_MODEL(chip_array[curr_chip].watchp) ,
                &iter, path ) )
    {
        return;
    }

    /*find the address of the selection*/
    gtk_tree_model_get( GTK_TREE_MODEL(chip_array[curr_chip].watchp), &iter,
            1, &addr, -1 );

    /*remove the watch point*/
    remove_watchpoint( curr_chip, addr );    
}

void break_popup_clear_break(GtkWidget *widget, gpointer data){
    GtkTreeIter iter;
    GtkTreePath *path;
    GtkTreeViewColumn *column;
    char *type;
    long addr;

    /*find out where the cursor is and highlight the row*/
    gtk_tree_view_get_cursor( (GTK_TREE_VIEW(chip_array[curr_chip].treeview_break)), 
        &path, &column );
    gtk_tree_view_row_activated ( GTK_TREE_VIEW( chip_array[curr_chip].treeview_break ),
         path, column );

    /*if the iter does not exist, then return*/
    if( !gtk_tree_model_get_iter( 
                GTK_TREE_MODEL(chip_array[curr_chip].breakp) ,
                &iter, path ) )
    {
        return;
    }

    /*find the type and address of the selection*/
    gtk_tree_model_get( GTK_TREE_MODEL(chip_array[curr_chip].breakp), &iter, 0, &type,
            2, &addr, -1 );

    /*remove the appropriate breakpoint*/
    if( !strncmp( type, "ASM", 3 ) )
    {
        remove_asm_breakpoint( curr_chip, addr );    
    }
    else
    {
        remove_source_breakpoint( curr_chip, addr );
    }
    
}

void source_popup_clear_break(GtkMenuItem *menuitem, gpointer user_data){
    GtkTextMark *mark;
    GtkTextIter iter;
    GtkTextBuffer *buff;
    char *cfile;
    int line;
    long addr;

    buff = gtk_text_view_get_buffer(GTK_TEXT_VIEW(chip_array[curr_chip].textview_source));
    mark = gtk_text_buffer_get_insert(buff);
    gtk_text_buffer_get_iter_at_mark(buff, &iter, mark);
    line = gtk_text_iter_get_line(&iter) + 1;
    cfile = cfile_index_to_cfile(curr_chip, gtk_option_menu_get_history(GTK_OPTION_MENU(chip_array[curr_chip].cfile_select)));
    addr = c_line_to_address(curr_chip, cfile, line);

    remove_source_breakpoint(curr_chip, addr);
}

void code_popup_clear_break(GtkMenuItem *menuitem, gpointer user_data){
    GtkTextMark *mark;
    GtkTextIter iter;
    int line;
    long addr;

    mark = gtk_text_buffer_get_insert(chip_array[curr_chip].code);
    gtk_text_buffer_get_iter_at_mark(chip_array[curr_chip].code, &iter, mark);
    line = gtk_text_iter_get_line(&iter);
    addr = text_line_to_address(curr_chip, line);

    remove_asm_breakpoint(curr_chip, addr);
}

gboolean cursor_to_click(GtkTextView *view, GdkEventButton *event, gpointer data) {
    if (event->button == 3) {
        int x, y;
        GtkTextIter iter;

        gtk_text_view_window_to_buffer_coords(view, 
                GTK_TEXT_WINDOW_TEXT, 
                event->x, event->y,
                &x, &y);
        gtk_text_view_get_iter_at_location(view, &iter, x, y);
        gtk_text_buffer_place_cursor(gtk_text_view_get_buffer(view), &iter);
    }
    return FALSE;
}

gboolean break_window_popup(GtkWidget *widget, GdkEventButton *event, gpointer data){
    if (event->button == 3) {
        gtk_menu_popup( GTK_MENU( chip_array[curr_chip].treeview_break_popup ), NULL, NULL,
            NULL, NULL, event->button, event->time );
    }
    return FALSE;
}    

gboolean watch_window_popup(GtkWidget *widget, GdkEventButton *event, gpointer data){
    if (event->button == 3) {
        gtk_menu_popup( GTK_MENU( chip_array[curr_chip].treeview_watch_popup ), NULL, NULL,
            NULL, NULL, event->button, event->time );
    }
    return FALSE;
}    

void source_window_popup(GtkTextView *textview, GtkMenu *arg1, gpointer user_data){
    GtkWidget *add_break;
    GtkWidget *clear_break;
    GtkWidget *icon;
    GtkTextMark *mark;
    GtkTextIter iter;
    GtkTextBuffer *buff;
    GtkWidget *divider;
    char *cfile;
    int line;
    long addr;
    int bpoint;

    /* calculate address to see if we have a breakpoint there or not */
    buff = gtk_text_view_get_buffer(GTK_TEXT_VIEW(chip_array[curr_chip].textview_source));
    mark = gtk_text_buffer_get_insert(buff);
    gtk_text_buffer_get_iter_at_mark(buff, &iter, mark);
    line = gtk_text_iter_get_line(&iter) + 1;
    cfile = cfile_index_to_cfile(curr_chip, gtk_option_menu_get_history(GTK_OPTION_MENU(chip_array[curr_chip].cfile_select)));
    addr = c_line_to_address(curr_chip, cfile, line);
    bpoint = is_source_breakpoint(curr_chip, addr);

    /* divider */
    divider = gtk_separator_menu_item_new();
    gtk_widget_show(divider);
    gtk_menu_shell_prepend(GTK_MENU_SHELL(arg1), divider);
    gtk_widget_set_sensitive(divider, FALSE);

    /* clear breakpoint */
    icon = gtk_image_new_from_stock ("gtk-remove", GTK_ICON_SIZE_MENU);
    clear_break = gtk_image_menu_item_new_with_label("Clear Breakpoint");
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(clear_break), icon);
    gtk_widget_show_all(clear_break);
    gtk_menu_shell_prepend(GTK_MENU_SHELL(arg1), clear_break);
    g_signal_connect(G_OBJECT(clear_break), "activate", G_CALLBACK(source_popup_clear_break), NULL);
    if(!bpoint || addr == -1){
        gtk_widget_set_sensitive(clear_break, FALSE);
    }

    /*  add breakpoint */
    icon = gtk_image_new_from_stock ("gtk-add", GTK_ICON_SIZE_MENU);
    add_break = gtk_image_menu_item_new_with_label("Add Breakpoint");
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(add_break), icon);
    gtk_widget_show_all(add_break);
    gtk_menu_shell_prepend(GTK_MENU_SHELL(arg1), add_break);
    g_signal_connect(G_OBJECT(add_break), "activate", G_CALLBACK(source_popup_add_break), NULL);
    if(bpoint || addr == -1){
        gtk_widget_set_sensitive(add_break, FALSE);
    }
}

void code_window_popup(GtkTextView *textview, GtkMenu *arg1, gpointer user_data){
    GtkWidget *add_break;
    GtkWidget *clear_break;
    GtkWidget *divider;
    GtkWidget *icon;
    GtkTextMark *mark;
    GtkTextIter iter;
    int line;
    long addr;
    int bpoint;

    mark = gtk_text_buffer_get_insert(chip_array[curr_chip].code);
    gtk_text_buffer_get_iter_at_mark(chip_array[curr_chip].code, &iter, mark);
    line = gtk_text_iter_get_line(&iter);
    addr = text_line_to_address(curr_chip, line);
    bpoint = is_asm_breakpoint(curr_chip, addr);

    /* divider */
    divider = gtk_separator_menu_item_new();
    gtk_widget_show(divider);
    gtk_menu_shell_prepend(GTK_MENU_SHELL(arg1), divider);
    gtk_widget_set_sensitive(divider, FALSE);

    /* clear breakpoint */
    icon = gtk_image_new_from_stock ("gtk-remove", GTK_ICON_SIZE_MENU);
    clear_break = gtk_image_menu_item_new_with_label("Clear Breakpoint");
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(clear_break), icon);
    gtk_widget_show_all(clear_break);
    gtk_menu_shell_prepend(GTK_MENU_SHELL(arg1), clear_break);
    g_signal_connect(G_OBJECT(clear_break), "activate", G_CALLBACK(code_popup_clear_break), NULL);
    if(!bpoint || addr == -1){
        gtk_widget_set_sensitive(clear_break, FALSE);
    }

    /*  add breakpoint */
    icon = gtk_image_new_from_stock ("gtk-add", GTK_ICON_SIZE_MENU);
    add_break = gtk_image_menu_item_new_with_label("Add Breakpoint");
    gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(add_break), icon);
    gtk_widget_show_all(add_break);
    gtk_menu_shell_prepend(GTK_MENU_SHELL(arg1), add_break);
    g_signal_connect(G_OBJECT(add_break), "activate", G_CALLBACK(code_popup_add_break), NULL);
    if(bpoint || addr == -1){
        gtk_widget_set_sensitive(add_break, FALSE);
    }

}

/* user has selected a different C file for display */
void select_cfile(GtkWidget *widget, gpointer data){
    GtkTextBuffer *buff;
    int chip_index;
    chip_data *chip = (chip_data *)data;

    chip_index = gtk_option_menu_get_history(GTK_OPTION_MENU(chip->cfile_select));

    buff = cfile_index_to_textbuffer(chip->id, chip_index);
    gtk_text_view_set_buffer(GTK_TEXT_VIEW(chip->textview_source), buff);
}

void global_popup_clear_watch()
{
    GtkTreeIter iter;
    GtkTreeSelection *select;
    GtkTreeModel *tm;

    GtkTreeView *treev = GTK_TREE_VIEW( chip_array[curr_chip].treeview_globals );

    int id;
    struct tuple_global_addr *glob;

    select = gtk_tree_view_get_selection( treev );
    gtk_tree_selection_get_selected( select, &tm, &iter );

    gtk_tree_model_get( tm, &iter, 3, &id, -1 );

    glob = id_to_global_data( curr_chip, id );
    if( glob )
    {
        if(glob->td != NULL)
        {
            remove_watchpoint_range( curr_chip, glob->addr, glob->td->byte_length );
        }
        else
        {
            remove_watchpoint_range( curr_chip, glob->addr, 4 );
        }       
    }

}

void global_popup_add_watch()
{
    GtkTreeIter iter;
    GtkTreeSelection *select;
    GtkTreeModel *tm;

    GtkTreeView *treev = GTK_TREE_VIEW( chip_array[curr_chip].treeview_globals );

    int id;
    struct tuple_global_addr *glob;

    select = gtk_tree_view_get_selection( treev );
    gtk_tree_selection_get_selected( select, &tm, &iter );

    gtk_tree_model_get( tm, &iter, 3, &id, -1 );

    glob = id_to_global_data( curr_chip, id );
    if( glob )
    {
        if(glob->td != NULL)
        {
            add_watchpoint_range( curr_chip, glob->addr, glob->td->byte_length, glob->name );
        }
        else
        {
            add_watchpoint_range( curr_chip, glob->addr, 4, glob->name );
        }       
    }
}

gboolean click_globals(GtkWidget *tv, GdkEventButton *event, gpointer null)
{

    GtkTreeView *treev = GTK_TREE_VIEW( chip_array[curr_chip].treeview_globals );

    if ( event->type != GDK_2BUTTON_PRESS )
    {    
        if ( event->button == 3 )
        {
            GtkWidget *global_watch;
            GtkWidget *icon;
            GtkWidget *clear_watch, *add_watch;

            GtkTreeIter iter;
            GtkTreePath *path;

            int id;
            struct tuple_global_addr *glob;

            /* try to get path to tre node from event location */
            if(!gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW(tv), event->x, event->y, &path, NULL, NULL, NULL)){
                return FALSE;
            }
            
            /*if the iter does not exist, then return*/
            if( !gtk_tree_model_get_iter( 
                        GTK_TREE_MODEL(chip_array[curr_chip].globals) ,
                        &iter, path ) )
            {
                return FALSE;
            }
            
            gtk_tree_model_get( GTK_TREE_MODEL(chip_array[curr_chip].globals), &iter, 3, &id, -1 );

            glob = id_to_global_data( curr_chip, id );
            
            global_watch = gtk_menu_new();

            /* clear watchpoint */
            icon = gtk_image_new_from_stock ("gtk-remove", GTK_ICON_SIZE_MENU);
            clear_watch = gtk_image_menu_item_new_with_label("Clear Watchpoint");
            gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(clear_watch), icon);
            gtk_widget_show_all(clear_watch);
            gtk_menu_shell_prepend(GTK_MENU_SHELL(global_watch), clear_watch);
            g_signal_connect(G_OBJECT(clear_watch), "activate", G_CALLBACK(global_popup_clear_watch), NULL);
            if(!is_watchpoint( curr_chip, glob->addr )){
                gtk_widget_set_sensitive(clear_watch, FALSE);
            }

            /*  add watchpoint */
            icon = gtk_image_new_from_stock ("gtk-add", GTK_ICON_SIZE_MENU);
            add_watch = gtk_image_menu_item_new_with_label("Add Watchpoint");
            gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(add_watch), icon);
            gtk_widget_show_all(add_watch);
            gtk_menu_shell_prepend(GTK_MENU_SHELL(global_watch), add_watch);
            g_signal_connect(G_OBJECT(add_watch), "activate", G_CALLBACK(global_popup_add_watch), NULL);
            if(is_watchpoint( curr_chip, glob->addr )){
                gtk_widget_set_sensitive(add_watch, FALSE);
            }
            
            gtk_menu_popup( GTK_MENU( global_watch ), NULL, NULL,
                    NULL, NULL, event->button, event->time );
        }
        return FALSE;
    }

    if ( event->button == 1 )
    {
        GtkTreeSelection *select;
        GtkTreeModel *tm;
        GtkTreeIter iter;

        int id;
        struct tuple_global_addr *glob;

        select = gtk_tree_view_get_selection( treev );
        gtk_tree_selection_get_selected( select, &tm, &iter );

        gtk_tree_model_get( tm, &iter, 3, &id, -1 );

        glob = id_to_global_data( curr_chip, id );
        if ( glob )
        {
            glob->type_style = 1 - glob->type_style;
            render_global( curr_chip, glob );
        }
    }

    return FALSE;
}

gboolean click_locals(GtkWidget *tv, GdkEventButton *event, gpointer null)
{
    GtkTreeView *treev = GTK_TREE_VIEW( tv );

    if ( event->type != GDK_2BUTTON_PRESS )
        return FALSE;

    if ( event->button == 1 )
    {
        GtkTreeSelection *select;
        GtkTreeModel *tm;
        GtkTreeIter iter;

        int id;
        struct local *loc;

        select = gtk_tree_view_get_selection( treev );
        gtk_tree_selection_get_selected( select, &tm, &iter );

        gtk_tree_model_get( tm, &iter, 3, &id, -1 );

        loc = id_to_local_data( curr_chip, id );
        if ( loc )
        {
            loc->type_style = 1 - loc->type_style;
            render_local( curr_chip, loc, &iter );
        }
    }

    return FALSE;
}

